/*
 * The SmartWeb Framework Maven Plugin
 * Copyright (C) 2004-2006
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * For further informations on the SmartWeb Framework Maven Plugin please visit
 *
 *                        http://smartweb.sourceforge.net/maven
 */
package net.smartlab.web.maven;

import java.io.File;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import net.smartlab.web.BusinessObject;
import net.smartlab.web.generator.Main;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.logging.Log;
import org.codehaus.plexus.compiler.util.scan.InclusionScanException;
import org.codehaus.plexus.compiler.util.scan.SimpleSourceInclusionScanner;
import org.codehaus.plexus.compiler.util.scan.SourceInclusionScanner;
import org.codehaus.plexus.compiler.util.scan.mapping.SuffixMapping;

/**
 * @author gcontartese
 */
public abstract class AbstractGenerator extends AbstractMojo {

	protected Log logger = super.getLog();

	/**
	 * The directory for compiled classes.
	 * 
	 * @parameter expression="${project.build.outputDirectory}"
	 * @required
	 * @readonly
	 */
	private File outputDirectory;

	/**
	 * Project classpath.
	 * 
	 * @parameter expression="${project.compileClasspathElements}"
	 * @required
	 * @readonly
	 */
	private List classpathElements;

	/**
	 * Directory containing the sources.
	 * 
	 * @parameter expression="${project.build.sourceDirectory}"
	 * @required
	 */
	private File sourceDirectory;

	/**
	 * Package of generate classes
	 * 
	 * @parameter expression="${package}"
	 * @required
	 */
	private String workPackage;

	/**
	 * Bundle of generate classes
	 * 
	 * @parameter expression="${bundle}" default-value=generated;
	 */
	private String bundle;
	
	/**
     * Expression to generate or not default unit tests case for each phase 
     * 
     * @parameter expression="${test}" default-value=true;
     */
    private String test;

	/**
	 * A list of inclusion filters for the compiler.
	 * 
	 * @parameter
	 */
	protected Set includes = new HashSet();

	/**
	 * A list of exclusion filters for the compiler.
	 * 
	 * @parameter
	 */
	protected Set excludes = new HashSet();

	/**
	 * A list of BusinessObject to be processed.
	 */
	private Set classes = new HashSet();

	/**
	 * Base directory where all src are written to.
	 * 
	 * @parameter expression="${sourceOutputDirectory}"
	 *            default-value="${project.build.directory}/smartweb/src"
	 */
	private File sourceOutputDirectory;
	
	/**
     * Base directory where all test are written to.
     * 
     * @parameter expression="${testOutputDirectory}"
     *            default-value="${project.build.directory}/smartweb/test"
     */
    private File testOutputDirectory;

	/**
	 * Base directory where all web are written to.
	 * 
	 * @parameter expression="${warSourceOutputDirectory}"
	 *            default-value="${project.build.directory}/smartweb/web"
	 */
	private File warSourceOutputDirectory;

	/**
	 * String represents a list of full qualified name of BusinessObject
	 * separated by comma. It become usefull when you want to lunch the plugin
	 * for only a limited classes. NB the classes must exist and must be
	 * assignable from BusinessObject.
	 * 
	 * @parameter expression="${classes}" default-value="empty"
	 */
	private String classList;

	/**
	 * Number of <code>BusinessOject</code> skipped during build
	 */
	private int skipped;

	/**
	 * Number of <code>BusinessOject</code> processed
	 */
	private int processed;


	/**
	 * @see org.apache.maven.plugin.AbstractMojo#execute()
	 */
	public void execute() throws MojoExecutionException, MojoFailureException {
		logger.info("Starting generation files for package .. " + workPackage);
		if (logger.isDebugEnabled()) {
			logger.debug("Source directory : " + sourceDirectory.toString());
			logger.debug("ClasspathElement : " + classpathElements.toString().replace(",", "\n"));
			logger.debug("Classlist :" + classList);
			for (Iterator it = classpathElements.iterator(); it.hasNext();) {
				String s = (String)it.next();
				logger.debug(" " + s);
			}
			logger.debug("OutuputDirectory : " + outputDirectory);
		}
		// load classes from the specified package
		Booter booter = new Booter();
		try {
			ClassLoader loader = booter.getClassLoader(getClasses(getSourceInclusionScanner("class")),
					classpathElements);
			classes = getClassElements(BusinessObject.class, outputDirectory, loader);
			Main main = new Main();
			if (classes != null && classes.size() == 0) {
				StringBuffer out = new StringBuffer();
				out.append("\n");
				out.append("\t ------------------------------------------------------------------------\n");
				out.append("\t NO BUSINESS OBJECT FOUND FOR PACKAGE :" + workPackage + "\n");
				out.append("\t Extends your classes using net.smartlab.web.BusinessObject \n");
				out.append("\t ------------------------------------------------------------------------\n");
				throw new MojoExecutionException(out.toString());
			}
			logger.info("------------------------------------------------------------------------");
			logger.info("BusinessObject to process :");
			StringBuffer bo = new StringBuffer();
			bo.append("\n");
			for (Iterator iterator = classes.iterator(); iterator.hasNext();) {
				Class type = (Class)iterator.next();
				bo.append("       " + type.getName() + "\n");
			}
			logger.info(bo.toString());
			main.setClasses(classes);
			main.setLog(logger);
			logger.info("------------------------------------------------------------------------");
			logger.info("Processing start... ");
			System.setProperty("sourceOutputDirectory", sourceOutputDirectory.getAbsolutePath());
			System.setProperty("warSourceOutputDirectory", warSourceOutputDirectory.getAbsolutePath());
			System.setProperty("testOutputDirectory", testOutputDirectory.getAbsolutePath());
			System.setProperty("package", workPackage);
			System.setProperty("bundle", bundle);
			System.setProperty("test", test);
			executeFromMain(main);
			logger.info("FINISH - Processed " + processed + " Skipped " + skipped);
		} catch (Exception e) {
			throw new MojoExecutionException("Failed to load BusinessObjects from ClassLoader", e);
		}
	}

	/**
	 * Execute the specific methods from main. Any subclasses must implements
	 * this method to run one or more Main methods after a simple configuration.
	 * 
	 * @param main The main to use to run the methods
	 * @throws MojoExecutionException It will be thrown if any error happened
	 *         during build process
	 */
	protected abstract void executeFromMain(Main main) throws MojoExecutionException;

	/**
	 * @return a Set of path representing classes
	 */
	private Set getClasses(SourceInclusionScanner scanner) {
		try {
			return scanner.getIncludedSources(outputDirectory, null);
		} catch (InclusionScanException e) {
			logger.warn("Failed to add classes from " + outputDirectory);
			return null;
		}
	}

	/**
	 * TODO documentation
	 * 
	 * @param inputFileEnding
	 * @return
	 */
	protected SourceInclusionScanner getSourceInclusionScanner(String inputFileEnding) {
		SourceInclusionScanner scanner = null;
		if (includes.isEmpty() && excludes.isEmpty()) {
			includes = Collections.singleton("**/*." + inputFileEnding);
			scanner = new SimpleSourceInclusionScanner(includes, Collections.EMPTY_SET);
		} else {
			if (includes.isEmpty()) {
				includes.add("**/*." + inputFileEnding);
			}
			scanner = new SimpleSourceInclusionScanner(includes, excludes);
		}
		// the following is needed by SimpleSourceInclusionScanner implementation
		scanner.addSourceMapping(new SuffixMapping("", ""));
		return scanner;
	}

	/**
	 * TODO documentation
	 * 
	 * @param type
	 * @param root
	 * @param loader
	 * @return
	 */
	protected Set getClassElements(Class type, File root, ClassLoader loader) {
		Set lista = getClasses(getSourceInclusionScanner("class"));
		Iterator iter = lista.iterator();
		Set classes = new LinkedHashSet();
		// Thread.currentThread().setContextClassLoader(loader);
		while (iter.hasNext()) {
			String clazz = getQualifiedName((File)iter.next(), root);
			Class to = null;
			try {
				to = loader.loadClass(clazz);
				type = loader.loadClass("net.smartlab.web.BusinessObject");
				if (type != null && type.isAssignableFrom(to) && !to.getName().equals(type.getName())) {
					if (to.getPackage().getName().equals(workPackage)) {
						classes.add(to);
						processed++;
					} else {
						logger.warn("Skipping " + to.getName() + " as " + type.getName() + " not in the right package");
						skipped++;
					}
				} else {
					logger.debug("Ignoring " + clazz + " as non inheriting from net.smartlab.web.BusinessObject");
				}
			} catch (Throwable e) {
				logger.error("Failed to load resource " + clazz + " resource will be skipped", e);
			}
		}
		return classes;
	}

	/**
	 * TODO documentation
	 * 
	 * @param file
	 * @param rootOutputDir
	 * @return
	 */
	private String getQualifiedName(File file, File rootOutputDir) {
		int length = (file.getAbsolutePath()).length();
		int rootLenght = rootOutputDir.getAbsolutePath().length();
		int start = length - (length - rootLenght) + 1;
		String name = (file.getAbsolutePath()).substring(start);
		name = name.substring(0, name.indexOf(".class"));
		String clazz = name.replace(File.separatorChar, '.');
		return clazz;
	}
}
